/***************************************************************************
 * File:        ./InputHandler.cpp
 * Author:      Ronni Grapenthin, NORVULK & HU-BERLIN
 * Created:     20.02.2007
 * Licence:     GPLv2
 *
 * #########################################################################
 *
 * CrusDe, simulation framework for crustal deformation studies
 * Copyright (C) 2007 Ronni Grapenthin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * -------------------------------------------------------------------------
 * Description: Reads commandline:
 *			- check whether we're dealing with an XML file
 *			- if not, check whether we're having some params given
 ****************************************************************************/

#include "SimulationCore.h"
#include "InputHandler.h"

#include <stdlib.h>	/*atof*/
#include <unistd.h>	/*getopt*/
#include <string.h>	/*strtok*/
#include <ctype.h>	/*isprint*/

#include <string>	
#include <map>
#include <stdexcept>

#include "PluginManager.h"

XERCES_CPP_NAMESPACE_USE
using namespace std;

/*------------------------------------------------------------------------------*/
/* USAGE INFO									*/
/*------------------------------------------------------------------------------*/
static void usage()
{
    cerr << "\nUsage:\n"
            "    crusde <XML file | <guiOptions> >\n\n"
            "This program runs an experiment that is parameterized by an XML file, \n"
	    "or starts a gui for either the experiment manager or the plug-in manager.\n"
            "guiOptions:\n"
	    "    -P          run Plug-in manager.\n" 
	    "    -M          run Experiment manager.\n" 
/*           "xmlOptions:\n"
            "    -l          Indicate the input file is a List File that has a list of xml files.\n"
            "                Default to off (Input file is an XML file).\n"
            "    -v=xxx      Validation scheme [always | never | auto*].\n"
            "    -n          Enable namespace processing. Defaults to off.\n"
            "    -s          Enable schema processing. Defaults to off.\n"
            "    -f          Enable full schema constraint checking. Defaults to off.\n"
            "    -locale=ll_CC specify the locale, default: en_US.\n"
            "    -p          Print out names of elements and attributes encountered.\n"
	    "    -?, -h      Show this help.\n\n"
	    "configOptions:\n"
	    "    -u          Update Plugin Database (look for new plugins in file system,\n"
	    "                delete plugins from database that are not found in file system).\n\n"
	    "experimentOptions:\n"
	    "    -Ga/b/c/... Parameter values for the Green's function plug-in.**\n" 
	    "    -La/b/c/... Parameter values for the Load function plug-in.**\n"
	    "    -Ka/b/c/... \n"
	    "    -Rwest/east/south/north\n"
	    "                Determines the region in which the modelling shall take plae\n"
	    "    -g=xxx      Green's function plug-in.\n" 
	    "    -l=xxx      Load function plug-in.\n" 
	    "    -k=xxx      Kernel plug-in. Defaults to 'fast-conv'.\n" 
            "  * = Default if not provided explicitly.\n"
	    " ** = Parameter sequence is defined within the scope of each plug-in.\n"
	    "      Usage is found elsewhere.\n"
*/         << endl;
}

InputHandler::InputHandler(int Argc, char *Argv[]):
	XMLHandler("dummy"),
	argc(Argc),
	isXML(false),
	doList(false)
{
	argv = Argv;
	
	//if input is xml, built dom ... else wait
	Debug("%s, line: %d, InputHandler built for: %s ", __FILE__, __LINE__, argv[0]);

}

InputHandler::~InputHandler()
{
	Debug("%s, line: %d, InputHandler destroyed.", __FILE__, __LINE__);
}

void InputHandler::init()
{

	int n(1);
	
	if (n == argc){
		usage();
		exit(2);
	}
	
	/* first level argument parsing to determine basic switches		*/
	/* rest will be done in input handler, where it will be possible	*/
	/* to set the plugin parameters right away.				*/
	while (n < argc )
	{
		// break at param that has no dash and following arguments contains ".xml" which means a xml-file is to be processed
        	if (argv[n][0] != '-' && (string(argv[n]).find(".xml", 0) != string::npos) ){
			isXML = true;
            		break;
		}

        	// help request
        	if (!strncmp(argv[n], "-?", 2) || !strncmp(argv[n], "-h", 2) )
        	{
            		usage();
            		exit(2);
        	}
		// input file holds a list of xml files
/*		else if (!strncmp(argv[n], "-l", 2) || !strncmp(argv[n], "-L", 2))
		{
			Debug("%s, line: %d, input file is a list file.", __FILE__, __LINE__);
			isXML = true;
			doList = true;
		}
*/		
		++n;
	}
	
	/* in case of xml processing there should be only 1 argument left, 	*/
	/* if experiment is set via command line, none should be left.		*/
	if ( (isXML && n != argc - 1) /*||  (!isXML && n != argc)*/ )
	{
		usage();
		exit(2);
	}

	if(isXML)
	{
		xmlFile = string(argv[n]);
	}	

	/* decide whether to process an xml input file or to have a closer look	*/
	/* at the command line arguments and built a DOM from there.		*/	
	if(isXML)
	{	
		try
		{
			readXML();
		}
		catch ( std::runtime_error e)
		{
			cerr << "Error in InitXML : " << e.what() << endl;
		}
	}
	else
	{
		readCommandline();
	}

}

void InputHandler::readCommandline()
{

Debug("%s, line: %d, read command line", __FILE__, __LINE__);

	char c;
	char delims[]=":";		/* delimiter in command line strings	*/
	char *result=NULL;            	/* dummy used in string processing	*/
	DOMElement *node;
	
	/* init Document Object Model where command line args will be saved to */
	initDOM();
	
	/* process command line option and set variables*/     
        while ((c = getopt (argc, argv, "G:g:L:l:K:k:pPmM")) != -1){

        /* get first parameter, used in all cases */
	   result = strtok( optarg, delims );

	/* figure out where we at in the parameter space */
           switch (c)
           {
               case 'G'://get GREEN Options
			node = greenElem;
			break;

	       case 'g'://get GREEN Plugin
	       		addPluginToDOM(greenElem, result);
			result = NULL;
			break;

               case 'L'://get LOAD Options
			node = loadElem;
			break;

	       case 'l'://get LOAD Plugin
	       		addPluginToDOM(loadElem, result);
			result = NULL;
			break;

               case 'K'://get KERNEL Options
			node = kernelElem;
			break;

	       case 'k'://get KERNEL Plugin
	       		addPluginToDOM(kernelElem, result);
			result = NULL;
			break;
	       case 'm': //start experiment manager
	       case 'M': //start experiment manager
	       		SimulationCore::instance()->runExperimentManager();
			break;
	       case 'p': //start plugin manager
	       case 'P': //start plugin manager
	       		SimulationCore::instance()->runPluginManager();
			break;
               default:
                      	abort ();
           }

	/* read parameters into dom */
	   int i(0);
	   char buffer[10];
	   
	   while(result != NULL)
	   {
		sprintf(buffer, "%d", i);
		addParamToDOM(node, buffer , result);
		result  = strtok( NULL, delims );
		++i;
           }
	}
}


void InputHandler::initParamsFromDOM()
{
	assert(doc);

	Debug("%s, line: %d, initParamsFromDOM().", __FILE__, __LINE__);

	DOMElement *root = doc->getDocumentElement();

	if(!isXML)
	{
		resolveParamNamesInDOM(root->getElementsByTagName(TAG_green.xmlStr())->item(0), SimulationCore::s_params.green);
		resolveParamNamesInDOM(root->getElementsByTagName(TAG_load.xmlStr())->item(0), SimulationCore::s_params.load);
		resolveParamNamesInDOM(root->getElementsByTagName(TAG_kernel.xmlStr())->item(0), SimulationCore::s_params.kernel);	
		resolveParamNamesInDOM(root->getElementsByTagName(TAG_postprocessor.xmlStr())->item(0), SimulationCore::s_params.postprocessor);	
		if(root->getElementsByTagName(TAG_loadhistory.xmlStr())->getLength() > 0)
			resolveParamNamesInDOM(root->getElementsByTagName(TAG_loadhistory.xmlStr())->item(0), SimulationCore::s_params.loadhistory);		
	}

	initParamsFromDOM(root->getElementsByTagName(TAG_green.xmlStr())->item(0), SimulationCore::s_params.green);
	initParamsFromDOM(root->getElementsByTagName(TAG_load.xmlStr())->item(0), SimulationCore::s_params.load);
	initParamsFromDOM(root->getElementsByTagName(TAG_kernel.xmlStr())->item(0), SimulationCore::s_params.kernel);
	initParamsFromDOM(root->getElementsByTagName(TAG_postprocessor.xmlStr())->item(0), SimulationCore::s_params.postprocessor);
	if(root->getElementsByTagName(TAG_loadhistory.xmlStr())->getLength() > 0)
		initParamsFromDOM(root->getElementsByTagName(TAG_loadhistory.xmlStr())->item(0), SimulationCore::s_params.loadhistory);
}

/**
 * set names of DOM nodes tht came from command line to something from the registry.
 */
void InputHandler::resolveParamNamesInDOM(DOMNode *node, multimap<string, double*> registry)
{
	assert(node);
	
	Debug("%s, line: %d, resolveParamNamesInDOM for node: %s.", __FILE__, __LINE__, XMLString::transcode(node->getNodeName()));
	
	DOMNode *child = node->getFirstChild();
	DOMNamedNodeMap *attributes = NULL;
	DOMNode *attr = NULL;
	multimap<string, double*>::iterator iter = registry.begin();
		
	while (child)
	{
		if( child->getNodeType() == DOMNode::ELEMENT_NODE)
	       	{			
			attributes = child->getAttributes();
			attr = attributes->getNamedItem(ATTR_name.xmlStr());
			
			char* name = XMLString::transcode(attr->getNodeValue());

			if( XMLString::compareIString(child->getNodeName(), TAG_parameter.xmlStr()) == 0 ){
				if(iter == registry.end())
				{	
					char* nodeName = XMLString::transcode(node->getNodeName());
					cerr << "Error: Too many Parameters for "<< nodeName << " in command line. See plugin documentation." << endl;
					XMLString::release(&nodeName);
					usage();
					exit(2);
				}
				/* upcast child to element node */
				
				DOMElement *elem = dynamic_cast<DOMElement*>(child);
				assert (elem!=NULL);
				
				/* free memory of old value */				
				char* value = XMLString::transcode(elem->getAttribute(ATTR_name.xmlStr()));
				XMLString::release(&value);
				/* set real value gotten from plugin */
				elem->setAttribute(ATTR_name.xmlStr(), XMLString::transcode(iter->first.c_str()));

				++iter;
			}

			/* free memory */
			XMLString::release(&name);
         	}
		
		child = child->getNextSibling();
	}

	if(iter != registry.end())
	{	
		char* nodeName = XMLString::transcode(node->getNodeName());
		cerr << "Warning: Too few Parameters for "<< nodeName << " in command line. Some might remain unset.\n"
			"We'll see what happends." << endl;
		XMLString::release(&nodeName);
	}
}	

void InputHandler::initParamsFromDOM(DOMNode *node, multimap<string, double*> registry)
{

	assert(node);
	
	Debug("%s, line: %d, initParamsFromDOM for node: %s.", __FILE__, __LINE__, XMLString::transcode(node->getNodeName()));
	
	DOMNode *child = node->getFirstChild();
	DOMNamedNodeMap *attributes = NULL;
	DOMNode *attr = NULL;
	
	multimap<string, double*>::iterator key_iter;
	pair< multimap<string, double*>::iterator, multimap<string, double*>::iterator > key_range;
	
	list<string> params_set;
	
	int count_keys(-1);
	unsigned int count_set(0);
	
	while (child)
	{
		if( child->getNodeType() == DOMNode::ELEMENT_NODE)
	       	{
			attributes = child->getAttributes();
			attr = attributes->getNamedItem(ATTR_name.xmlStr());
			
			char* name = XMLString::transcode(attr->getNodeValue());

			if( XMLString::compareIString(child->getNodeName(), TAG_parameter.xmlStr()) == 0 )
			{			
				//get number of keys that equal name
				count_keys = static_cast<int>( registry.count(string(name)) );
				//if there is anything in the registry ... 
				if( count_keys > 0 )
				{		
					//equal_range gives two results: an iterator to the first and last element with key==name
					key_range = registry.equal_range(string(name));
					//all keys have values that are adresses of double variables in the repsective plugins
					//each of those variables now gets a value assigned. the same value.
					for ( key_iter=key_range.first; key_iter != key_range.second; ++key_iter)
					{
						//get the value from the DOM
						StrXML value(attributes->getNamedItem(ATTR_value.xmlStr())->getNodeValue());
						//write it into the variable that's strored at key_iter->second
						*(key_iter->second) = static_cast<double>( atof(value.cStr()) );
						++count_set;
					}
					//memorize the key that was set only once
					params_set.push_back(key_range.first->first);
				}
				else //if not found there might be a spelling mistake either in input, or plugin, or both :)
				{
					StrXML paramName(attributes->getNamedItem(ATTR_name.xmlStr())->getNodeValue());
					cerr << "Warning: Parameter "<< paramName.cppStr() << " (coming from the experiment definition) not registered! Misspelled in XML file or Plug-in definition?"
					     << endl;
				}
			}

         	}
		
		child = child->getNextSibling();
	}
	
	if(count_set<registry.size())
	{
		cerr<<"Error: Parameters that remain uninitialized: "<<endl;
		multimap<string, double*>::iterator map_iter = registry.begin();
		list<string>::iterator found_iter;		
		
		while (map_iter != registry.end())
		{
			found_iter = params_set.begin();
			//as long as we're not at the end and could not find the parameter in the list of set 
			//parameters, continue looking.
			while( found_iter != params_set.end() && (*found_iter).compare(map_iter->first) != 0  ) {
   				++found_iter;
			} 	
			//if we're at the end, the parameter was not in out list ... tell the user.			
			if(found_iter==params_set.end())
			{
				cerr<< "\t" << map_iter->first << endl;
			}
			
			//get to the next unique value in the registry ... avoid printing multiple keys twice.
			map_iter = ( registry.equal_range(map_iter->first) ).second;
		}
		cerr<<"Aborting."<<endl;
		exit(2);
	}
		
}	



void InputHandler::initDOM()
{

	DOMImplementation* impl =  DOMImplementationRegistry::getDOMImplementation(XMLString::transcode("Core"));

	if (impl != NULL)
	{
		try
		{
			doc = impl->createDocument(
				0,                    // root element namespace URI.
				TAG_experiment.xmlStr(),       // root element name
				0);                   // document type object (DTD).

			//get root, i.e. experiment, element
			rootElem = doc->getDocumentElement();

			// create major elements of an experiment
			greenElem = doc->createElement(TAG_green.xmlStr());
			loadElem = doc->createElement(TAG_load.xmlStr());
			kernelElem = doc->createElement(TAG_kernel.xmlStr());

			//append those to experiement (root)
			rootElem->appendChild(greenElem);
			rootElem->appendChild(loadElem);
			rootElem->appendChild(kernelElem);

		}
		catch (const OutOfMemoryException&)
		{
			XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
		}
		catch (const DOMException& e)
		{
			XERCES_STD_QUALIFIER cerr << "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;
		}
		catch (...)
		{
			XERCES_STD_QUALIFIER cerr << "An error occurred creating the document" << XERCES_STD_QUALIFIER endl;
		}
	}  // (impl != NULL)
	else
	{
			XERCES_STD_QUALIFIER cerr << "Requested implementation is not supported" << XERCES_STD_QUALIFIER endl;
	}
}

/**
 * this function adds a parameter tag with the attributes name and value to node:
 * 
 *    <node>
 *        <parameter name='name' value='value' /> <!-- this is added -->
 *    </node>
 */
void InputHandler::addParamToDOM(DOMElement *node, string name, string value)
{
Debug("%s, line: %d, adding %s=%s to %s", __FILE__, __LINE__, name.c_str(), value.c_str(), XMLString::transcode(node->getNodeName()));

	try
	{
		DOMElement*  parameter = doc->createElement(TAG_parameter.xmlStr());
		node->appendChild(parameter);
		parameter->setAttribute(ATTR_name.xmlStr(), XMLString::transcode(name.c_str()));
		parameter->setAttribute(ATTR_value.xmlStr(), XMLString::transcode(value.c_str()));
	}
	catch (const OutOfMemoryException&)
	{
		XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
	}
	catch (const DOMException& e)
	{
		XERCES_STD_QUALIFIER cerr << "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;
	}
	catch (...)
	{
		XERCES_STD_QUALIFIER cerr << "An error occurred creating the document" << XERCES_STD_QUALIFIER endl;
	}
}

/**
 * this function adds a plugin tag with the attribute name to node:
 * 
 *    <node>
 *        <plugin name='name' /> <!-- this is added -->
 *    </node>
 */
void InputHandler::addPluginToDOM(DOMElement *node, string name)
{
Debug("%s, line: %d, adding %s to %s", __FILE__, __LINE__, name.c_str(), XMLString::transcode(node->getNodeName()));
	
	try
	{
		DOMElement*  plugin = doc->createElement(TAG_plugin.xmlStr());
		node->appendChild(plugin);
		plugin->setAttribute(ATTR_name.xmlStr(), XMLString::transcode(name.c_str()));
	}
	catch (const OutOfMemoryException&)
	{
		XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
	}
	catch (const DOMException& e)
	{
		XERCES_STD_QUALIFIER cerr << "DOMException code is: " << e.code << XERCES_STD_QUALIFIER endl;
	}
	catch (...)
	{
		XERCES_STD_QUALIFIER cerr << "An error occurred creating the document" << XERCES_STD_QUALIFIER endl;
	}
}

string InputHandler::getAttributeValueByName(const XMLCh* elementNode, const XMLCh* attribute, const XMLCh* name)
{

Debug("%s, line: %d, InputHandler::getAttributeValueByName(%s) name = %s ", __FILE__, __LINE__, XMLString::transcode(elementNode),
XMLString::transcode(name));
	assert(doc);

	DOMElement *root = doc->getDocumentElement();

	DOMNode *child = root->getFirstChild();
	DOMNamedNodeMap *attributes = NULL;
		
	while (child)
	{
	if(child != NULL)
		if( child->getNodeType() == DOMNode::ELEMENT_NODE )
	       	{
			if(child->hasAttributes())
			{
				attributes = child->getAttributes();

				if( XMLString::compareIString( child->getNodeName(), elementNode) == 0 && 
				    XMLString::compareIString( attributes->getNamedItem(ATTR_name.xmlStr())->getNodeValue(), name) == 0 )
				{

					char *val = XMLString::transcode(attributes->getNamedItem(attribute)->getNodeValue());
					string value(val);
					XMLString::release(&val);
					return value;
				}
			}
		}

		child = child->getNextSibling();
	}
	
	return string();
}


/**
 * extract's the value of an attribute and returns it:
 *
 * <parentNode>
 *	<elementNode attribute="returnstring" />
 * </parentNode>
 *
 * the first parentNode found in the document is used. thus, it is expected to be unique.
 *
 * @param parentNode 
 * @param elementNode 
 * @param attribute 
 * @return 
 */
string InputHandler::getAttributeValue(const XMLCh* parentNode, const XMLCh* elementNode, const XMLCh* attribute)
{
Debug("%s, line: %d, InputHandler::getAttributeValue(%s) ", __FILE__, __LINE__, XMLString::transcode(elementNode));

	DOMElement *root = doc->getDocumentElement();
	DOMNodeList *node_list = root->getElementsByTagName(parentNode);
	
	/*if element does not exist, return emptry string*/
	if(node_list->getLength() == 0)
		return string();
		
	DOMNode *child = node_list->item(0)->getFirstChild();

	DOMNamedNodeMap *attributes = NULL;
		
	while (child)
	{
		if( child->getNodeType() == DOMNode::ELEMENT_NODE)
	       	{
			attributes = child->getAttributes();

			if( XMLString::compareIString(child->getNodeName(), elementNode) == 0 )
			{
				char *val = XMLString::transcode(attributes->getNamedItem(attribute)->getNodeValue());
				string value(val);
				XMLString::release(&val);
				return value;
			}
			
		}
		child = child->getNextSibling();
	}
	return string();
}

string InputHandler::getGreenPlugin()
{
	try{
		string name(getAttributeValue(TAG_green.xmlStr(), TAG_plugin.xmlStr(), ATTR_name.xmlStr()));
		return (SimulationCore::instance()->pluginManager())->getFilename(string("green"), name);	
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getGreenPlugin(string name)
{
	try{
		return SimulationCore::instance()->pluginManager()->getFilename("green", name);		
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getLoadPlugin()
{
	try{
		string name(getAttributeValue(TAG_load.xmlStr(), TAG_plugin.xmlStr(), ATTR_name.xmlStr()));
		return SimulationCore::instance()->pluginManager()->getFilename("load", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getLoadPlugin(string name)
{
	try{
		return SimulationCore::instance()->pluginManager()->getFilename("load", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getLoadHistoryPlugin()
{
	try{
		string name(getAttributeValue(TAG_loadhistory.xmlStr(), TAG_plugin.xmlStr(), ATTR_name.xmlStr()));
		if(!name.empty())
			return SimulationCore::instance()->pluginManager()->getFilename("load_history", name);
		else{
			return string();
		}
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getLoadHistoryPlugin(string name)
{
	try{
		return SimulationCore::instance()->pluginManager()->getFilename("load_history", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getKernelPlugin()
{
	try{
		string name(getAttributeValue(TAG_kernel.xmlStr(), TAG_plugin.xmlStr(), ATTR_name.xmlStr()));
		return SimulationCore::instance()->pluginManager()->getFilename("operator", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getKernelPlugin(string name)
{
	try{
		return SimulationCore::instance()->pluginManager()->getFilename("operator", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getOutputPlugin()
{
	try{
		string name(getAttributeValue(TAG_output.xmlStr(), TAG_plugin.xmlStr(), ATTR_name.xmlStr()));
		return SimulationCore::instance()->pluginManager()->getFilename("data_handler", name);
	}
	catch( ... ){
		throw;
	}

}

string InputHandler::getOutputPlugin(string name)
{
	try{
		return SimulationCore::instance()->pluginManager()->getFilename("data_handler", name);
	}
	catch( ... ){
		throw;
	}
}

string InputHandler::getFileName(string type)
{	
	StrXML t(type);
	return getAttributeValueByName(TAG_file.xmlStr(), ATTR_value.xmlStr(), t.xmlStr());
	
}

string InputHandler::getResultFileName()
{	
	StrXML r("result");
	return getAttributeValueByName(TAG_file.xmlStr(), ATTR_value.xmlStr(), r.xmlStr());
	
}

int InputHandler::getRegion(string direction)
{
	StrXML d(direction);
	return atoi(getAttributeValueByName(TAG_region.xmlStr(), ATTR_value.xmlStr(), d.xmlStr()).c_str());
}

int InputHandler::getGridSize()
{
	StrXML g("gridsize");
	return atoi(getAttributeValueByName(TAG_parameter.xmlStr(), ATTR_value.xmlStr(), g.xmlStr()).c_str());
}

int InputHandler::getTimeSteps()
{
	StrXML t("timesteps");
	string timesteps(getAttributeValueByName(TAG_parameter.xmlStr(), ATTR_value.xmlStr(), t.xmlStr()));
	if(!timesteps.empty())
		return atoi(timesteps.c_str());
	else
		return 1;
}

DOMNode* InputHandler::getExperiment()
{
	return doc->getDocumentElement();
}

list<string> InputHandler::getPostProcessorNameList()
{
	list<string> names;
	
	//get the postprocessor node 
	DOMElement *elem = dynamic_cast<DOMElement*>(getElementByTagName(TAG_postprocessor.xmlStr()));
	//check whether there is one
	if(!elem)
	{
		cerr<<"No postprocessors found."<<endl;
		return names;
	}
	//if so, grab all plugins it owns.
	DOMNodeList *plugins = elem->getElementsByTagName(TAG_plugin.xmlStr());
	unsigned int i(0);
	
	//if there are plugins ... 
	while(plugins && i<plugins->getLength())
	{
		if( plugins->item(i)->getNodeType() == DOMNode::ELEMENT_NODE)
		{			
			/* upcast child to element node */
			DOMElement *elem = dynamic_cast<DOMElement*>( plugins->item(i) );
			StrXML name(elem->getAttribute(ATTR_name.xmlStr()));
			//... add names to list for later lookup.
			names.push_back(name.cppStr());
		}
		++i;		
	}
	
	//return filled list
	return names;
}
