/***************************************************************************
 * File:        ./SimulationCore.cpp
 * Author:      Ronni Grapenthin, NORVULK & HU-BERLIN
 * Created:     20.02.2007
 * Licence:     GPLv2
 *
 * #########################################################################
 *
 * CrusDe, simulation framework for crustal deformation studies
 * Copyright (C) 2007 Ronni Grapenthin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 ****************************************************************************/

#include	"SimulationCore.h"
#include	"LoadHistoryPlugin.h"
#include	"DataOutPlugin.h"
#include	"ExperimentManager.h"
#include	"config.h"
#include	"Plugin.h"
#include	"ConvFactorPlugin.h"
#include	"InputHandler.h"
#include	"PluginManager.h"

#include 	<iostream>

#include 	<stdarg.h>
#include 	<math.h>


//! public struct keeping the command line arguments
s_parameters SimulationCore::s_params = {};

//! static Singleton initialisieren
SimulationCore* SimulationCore::pSimulationCore=0;

SimulationCore::SimulationCore(int argc, char** argv) throw (SeriousException) :
	com_port(NULL), exp_man(NULL), plugin_man(NULL), pl_kernel(NULL), 
	pl_out(NULL), pl_green(NULL), 
	pl_load(NULL), pl_history(NULL), pl_list_postprocessor(),
	x_west(0), x_east(0), y_south(0), y_north(0), 
	gridsize(0), modeltime(0), num_timesteps(1), 
	stepsize(1), quadrant(1), dimensions(0),
	x_index(-1), y_index(-1), z_index(-1),
	model_data(NULL), load_file(), obs_file(),
	requested_plugin_list(),
	root_dir("/")
{
	if(getenv("CRUSDE_HOME") == NULL)
	    	throw (SeriousException ( string("Error: Environment variable CRUSDE_HOME is not defined!") ) );

	root_dir = string(getenv("CRUSDE_HOME"));

	/*create integral parts*/
	com_port  = new InputHandler(argc, argv);
	exp_man   = new ExperimentManager(
				string(root_dir).append(DIR_SEP).append(EXPERIMENT_DIR).append(DIR_SEP).append(EXPERIMENT_DB).c_str()
			);
	plugin_man= new PluginManager(
				string(root_dir).append(DIR_SEP).append(PLUGIN_DIR).append(DIR_SEP).append(PLUGIN_DB).c_str()
			);

	
	/*create plugins*/
	pl_green  = new GreenPlugin("green");
	pl_load   = new LoadPlugin("load");
	pl_history= new LoadHistoryPlugin("load history");
	pl_out    = new DataOutPlugin("output");
	pl_kernel = new Plugin("modeling kernel");
}

SimulationCore::~SimulationCore()
{
	delete pl_kernel;
	delete pl_out;
	delete pl_history;
	delete pl_load;
	delete pl_green;
	delete com_port;
}

void SimulationCore::init() //throw INIT_EXCEPTION
{
	
    // ------------------
    // build DOM from command line args or XML input respectively
    // ------------------

	com_port->init();	
	exp_man->init();
	plugin_man->init();
    
    // ------------------
    // set setup variables ... important to set these BEFORE plugins get initialized!
    // ------------------
	load_file = string( com_port->getFileName("load") );
	obs_file = string( com_port->getFileName("observe") );
	
	x_west  =  com_port->getRegion("west");
	x_east  =  com_port->getRegion("east");
	y_south =  com_port->getRegion("south");
	y_north =  com_port->getRegion("north");
	
	gridsize = com_port->getGridSize();
	num_timesteps = com_port->getTimeSteps();	

    // ------------------
    // create all postprocessors
    // ------------------
	
	list<string> names = com_port->getPostProcessorNameList();
	list<string>::iterator post_iter = names.begin();
	
	while(post_iter != names.end()){
		Plugin *pl = new Plugin( (*post_iter).c_str() );
		//try loading the plugin	
		try
		{
			pl->load( plugin_man->getFilename("postprocess", (*post_iter)) );
		}
		catch(runtime_error e)
		{
			abort(e.what());
		}
		
		pl_list_postprocess.push_back( pl );
		++post_iter;
	}    
    
    // ------------------
    // Have plugins load their libaries
    // ------------------
	
	try
	{
		pl_green->load( com_port->getGreenPlugin() );
		pl_load->load( com_port->getLoadPlugin() );
		pl_history->load( com_port->getLoadHistoryPlugin() );
		pl_kernel->load( com_port->getKernelPlugin() );
		pl_out->load( com_port->getOutputPlugin() );
	}
	catch(runtime_error e)
	{
		abort(e.what());
	}
    
    // ------------------
    // Have plugins register their parameters
    // ------------------

	pl_green->registerParameter();
	pl_load->registerParameter();
	pl_history->registerParameter();
	pl_kernel->registerParameter();
	pl_out->registerParameter();
	//postprocessors
	pl_iter = pl_list_postprocess.begin();
	while(pl_iter != pl_list_postprocess.end()){
		(*pl_iter)->registerParameter();
		++pl_iter;
	}

    // ------------------
    // Have plugins allocate their output memory
    // ------------------

	pl_green->registerOutputFields();
	pl_iter = pl_list_postprocess.begin();
	while(pl_iter != pl_list_postprocess.end()){
		(*pl_iter)->registerOutputFields();
		++pl_iter;
	}

    // ------------------
    // Have plugins get access to other plugins
    // ------------------

    	pl_green->requestPlugins();
    	pl_load->requestPlugins();
	pl_history->requestPlugins();
    	pl_kernel->requestPlugins();
    	pl_out->requestPlugins();
	//postprocessors
	pl_iter = pl_list_postprocess.begin();
	while(pl_iter != pl_list_postprocess.end()){
		(*pl_iter)->requestPlugins();
		++pl_iter;
	}

    // ------------------
    // init Parameters that are now known to the registry
    // ------------------
	
	com_port->initParamsFromDOM();
	
    // ------------------
    // now the registered values are set and we can truly initialize the plugins 
    // ------------------

	pl_green->init();
	pl_load->init();
	pl_history->init();
	pl_kernel->init();
	pl_out->init();

	//postprocessors
	pl_iter = pl_list_postprocess.begin();
	while(pl_iter != pl_list_postprocess.end()){
		(*pl_iter)->init();
		++pl_iter;
	}
	
	//requested plugins
	pl_iter = requested_plugin_list.begin();
	while(pl_iter != requested_plugin_list.end()){
		(*pl_iter)->init();
		++pl_iter;
	}
}

void SimulationCore::terminate() //throw KILL_EXCEPTION
{
    // ------------------
    // Have plugins release their resources
    // ------------------

	pl_green->release();
	pl_load->release();
	pl_history->release();
	pl_kernel->release();
	pl_out->release();

	//postprocessors
	pl_iter = pl_list_postprocess.begin();
	while(pl_iter != pl_list_postprocess.end()){
		(*pl_iter)->release();
		++pl_iter;
	}
	pl_iter = requested_plugin_list.begin();
	while(pl_iter != requested_plugin_list.end()){
		(*pl_iter)->release();
		++pl_iter;
	}
    // ------------------
    // Have plugins unload their libraries
    // ------------------
	try
	{
		pl_green->unload();
		pl_load->unload();
		pl_history->unload();
		pl_kernel->unload();
		pl_out->unload();
		//postprocessors
		pl_iter = pl_list_postprocess.begin();
		while(pl_iter != pl_list_postprocess.end()){
			(*pl_iter)->unload();
			++pl_iter;
		}

		pl_iter = requested_plugin_list.begin();
		while(pl_iter != requested_plugin_list.end()){
			(*pl_iter)->unload();
			++pl_iter;
		}
	}
	catch(runtime_error e)
	{
		abort(e.what());
	}
    // ------------------
    // destroy singleton object
    // ------------------

	delete pSimulationCore;
}

/*! go!go!go!
 */
void SimulationCore::exec() //throw EXEC_EXCEPTION
{

cout << "\n-------------------------"<<endl;
	while(modeltime < num_timesteps)
	{
cout << "TIME: "<<modeltime<<endl;
		
cout << "operator starts ... "<<endl;
		pl_kernel->run();
cout << "postprocessors start ... "<<endl;
		//postprocessors
		pl_iter = pl_list_postprocess.begin();
		while(pl_iter != pl_list_postprocess.end()){
			(*pl_iter)->run();
			++pl_iter;
		}
cout << "result handler starts ... "<<endl;
		pl_out->run();
		
cout << "-------------------------"<<endl;
		++modeltime;
	}
	
	//if we're getting this far the experiment was a success ... memorize it!
	exp_man->addEntry(outFile(), com_port->getExperiment());	
}

/**
 * 
 * @param cat Plugin category
 * @return list of string IDs of registered Parameters
 */
list<string> SimulationCore::getRegisteredParameters(PluginCategory category)
{
	list<string> registeredParams;
	multimap<string, double*> p_map;
	
	if(category == LOAD_PLUGIN){
		p_map = SimulationCore::s_params.load;
	} 
	else if(category == KERNEL_PLUGIN){
		p_map = SimulationCore::s_params.kernel;
	} 
	else if(category == GREEN_PLUGIN){
		p_map = SimulationCore::s_params.green;
	} 
	else if(category == DATAOUT_PLUGIN){
		p_map = SimulationCore::s_params.datahandler;
	} 
	else if(category == POSTPROCESS_PLUGIN){
		p_map = SimulationCore::s_params.postprocessor;
	}
	else if(category == LOADHISTORY_PLUGIN){
		p_map = SimulationCore::s_params.loadhistory;
	} else {
		cerr<<"in SimulationCore::getRegisteredParameters: unknown category: "<<category<<endl;
	}

  	multimap<string , double*>::iterator iter = p_map.begin();
	
	while( iter != p_map.end()) 
	{
		registeredParams.push_back(iter->first);
		++iter;
 	}
	
	registeredParams.sort();
	registeredParams.unique();	//deletes consecutive duplicate elements, needs sort before,
			
	return registeredParams;	
}

/**
 * deletes all parameters of a specific category
 *
 * @param cat plugin category
 */
void SimulationCore::deleteRegistrees()
{
	
	SimulationCore::s_params.load.clear();
	SimulationCore::s_params.kernel.clear();
	SimulationCore::s_params.green.clear();
	SimulationCore::s_params.datahandler.clear();
	SimulationCore::s_params.postprocessor.clear();
	SimulationCore::s_params.loadhistory.clear();
}


list<string> SimulationCore::getRequestedNames()
{
	list<string> requestedNames;
	
	pl_iter = requested_plugin_list.begin();
	while(pl_iter != requested_plugin_list.end()){
		requestedNames.push_back((*pl_iter)->getName());
		++pl_iter;
	}
	
	return requestedNames;	
}

void SimulationCore::deleteRequests()
{
	pl_iter = requested_plugin_list.begin();
	while(pl_iter != requested_plugin_list.end()){
		(*pl_iter)->unload();
		delete(*pl_iter);
		++pl_iter;
	}
	
	requested_plugin_list.clear();
}


/*
 * a green plugin was requested for usage by another green plugin.
 * thus, it needs to be created, loaded, has it's parameters registerd.
 * it is added to the general plugin storage and this way kept in memory
 * and accessible to call init, unload, and other plugin functions at it.
 * it returns a pointer to the green's functions execution function.
 */
green_exec_function SimulationCore::addGreenPlugin(string plugin) throw (FileNotFound, runtime_error)
{
cout << "\nADDING GREEN PLUGIN TO requested_plugin_list"<<endl;
		
	try{
		GreenPlugin *pl = new GreenPlugin(plugin);
		string path(com_port->getGreenPlugin(plugin));
 	
		string error_string("Could not find plugin ");
	
		if(path.empty())
		{
			error_string.append("'").append(plugin).append("'\n");
			error_string.append("Not installed?\nMisspelled?\nCheck Database!\n");
			throw ( FileNotFound(error_string) );
		}
		
		pl->load( path );
		pl->registerParameter();
		pl->registerOutputFields();
		
		requested_plugin_list.push_back(pl);
		
		return pl->getArrayFunction();

	}
	catch( ... ){
		throw;
	}
}

/*
 * a load plugin was requested for usage by another green plugin.
 * thus, it needs to be created, loaded, has it's parameters registerd.
 * it is added to the general plugin storage and this way kept in memory
 * and accessible to call init, unload, and other plugin functions at it.
 * it returns a pointer to the load functions execution function.
 */
load_exec_function SimulationCore::addLoadPlugin(string plugin) throw (FileNotFound, runtime_error)
{
	try{
		LoadPlugin *pl = new LoadPlugin(plugin);
		string path(com_port->getLoadPlugin(plugin));
		string error_string("Could not find plugin ");
	
		if(path.empty())
		{
			error_string.append("'").append(plugin).append("'\n");
			error_string.append("Not installed?\nMisspelled?\nCheck Database!\n");
			throw ( FileNotFound(error_string) );
		}
	
		pl->load( path );
		pl->registerParameter();
		
		requested_plugin_list.push_back(pl);
		
		return pl->getValueFunction();

	}
	catch( ... ){
		throw;
	}
	
}

/*
 * a kernel plugin was requested for usage by another green plugin.
 * thus, it needs to be created, loaded, has it's parameters registerd.
 * it is added to the general plugin storage and this way kept in memory
 * and accessible to call init, unload, and other plugin functions at it.
 * it returns a pointer to the load functions execution function.
 */
run_function SimulationCore::addKernelPlugin(string plugin) throw (FileNotFound, runtime_error)
{
	try
	{
		Plugin *pl = new Plugin(plugin);
		string path(com_port->getKernelPlugin(plugin));
		string error_string("Could not find plugin ");
		
		if(path.empty())
		{
			error_string.append("'").append(plugin).append("'\n");
			error_string.append("Not installed?\nMisspelled?\nCheck Database!\n");
			throw ( FileNotFound(error_string) );
		}
			
		pl->load( path );
		pl->registerParameter();
		
		requested_plugin_list.push_back(pl);
		
		return pl->getRunFunction();
	}
	catch(...){
		throw;
	}
}

	
run_function SimulationCore::addPostprocessorPlugin(string plugin) throw (FileNotFound, runtime_error)
{
	try
	{
		Plugin *pl = new Plugin(plugin);
		string path( plugin_man->getFilename("postprocess", plugin));
	
		string error_string("Could not find plugin ");
		
		if(path.empty())
		{
			error_string.append("'").append(plugin).append("'\n");
			error_string.append("Not installed?\nMisspelled?\nCheck Database!\n");
			throw ( FileNotFound(error_string) );
		}
			
		pl->load( path );
		pl->registerParameter();
		
		requested_plugin_list.push_back(pl);
		
		return pl->getRunFunction();
	}
	catch(...){
		throw;
	}
}

loadhistory_exec_function SimulationCore::addLoadHistoryPlugin(string plugin) throw (FileNotFound, runtime_error)
{
	try{
		LoadHistoryPlugin *pl = new LoadHistoryPlugin(plugin);
		string path(com_port->getLoadHistoryPlugin(plugin));
		string error_string("Could not find plugin ");
	
		if(path.empty())
		{
			error_string.append("'").append(plugin).append("'\n");
			error_string.append("Not installed?\nMisspelled?\nCheck Database!\n");
			throw ( FileNotFound(error_string) );
		}
	
		pl->load( path );
		pl->registerParameter();
		
		requested_plugin_list.push_back(pl);
		
		return pl->getValueFunction();

	}
	catch( ... ){
		throw;
	}
	
}


void SimulationCore::registerOutputField(int *output_index, FieldName field)
{
	//check which field is gonna be set and whether or not it is already claimed by someone else:
	if(field == X_FIELD){
		if(x_index < 0){
			x_index = dimensions;
		}
		else{
			(*output_index) = x_index;
			return;
		}
	}
	else
	if(field == Y_FIELD){
		if(y_index < 0){
			y_index = dimensions;
		}
		else{
			(*output_index) = y_index;
			return;
		}
	}
	else
	if(field == Z_FIELD){
		if(z_index < 0){
			z_index = dimensions;
		}
		else{
			(*output_index) = z_index;
			return;
		}
	
	}
	else
	if(field == ADD_FIELD){
		//check whether the requesting address has already been assinged an output field
		map<int*,int>::iterator iter = add_field_map.find(output_index);
  		if( iter != add_field_map.end() ) 
		{
	    		cerr << "Warning: Seems like address '" << iter->first << "' has already requested memory. Is there one plugin trying to do its job twice? We will NOT give additional memory to that address."<< endl;
			return;
 		}
		else
		{
		//else memorize that address for later...	
 			add_field_map[output_index] = dimensions;
		}
	}
	else
	{
		abort("Trying to set Invalid field. Use X_FIELD, Y_FIELD, Z_FIELD, or ADD_FIELD only!");
	}

	//set field's position in output array
	(*output_index) = dimensions;
	
	//increase output dimension
	++dimensions;	
}

void SimulationCore::setModelData(double **result)
{ 
	model_data = result;
	pl_out->setModelData(result);
}

double** SimulationCore::getModelData()
{ 
	return model_data;
}

const char* SimulationCore::loadFile()
{
	return load_file.c_str();
}

const char* SimulationCore::outFile()
{
	string name(com_port->getResultFileName());
	
	if(name.empty())
		return "experiment.nc";
	else
		return name.c_str();
}

const char* SimulationCore::observeFile()
{	
	return obs_file.c_str();
}

int SimulationCore::sizeX()
{
	return static_cast<int>( ceil( static_cast<double>((x_east - x_west) / gridsize ) ) ) + 1 ;
}

int SimulationCore::sizeY()
{
	return static_cast<int>( ceil( static_cast<double>((y_north - y_south) / gridsize ) ) ) + 1 ;
}

int SimulationCore::sizeT()
{
	return num_timesteps;
}

int SimulationCore::minX()
{
	return x_west;
}

int SimulationCore::minY()
{
	return y_south;
}

int SimulationCore::gridSize()
{
	return gridsize;
}

int SimulationCore::stepSize()
{
	return stepsize;
}

int SimulationCore::modelTime()
{
	return modeltime;
}

int SimulationCore::xIndex()
{
	return x_index;
}

int SimulationCore::yIndex()
{
	return y_index;
}

int SimulationCore::zIndex()
{
 	return z_index;
}

int SimulationCore::getQuadrant()
{
	return quadrant;
}

void SimulationCore::setQuadrant(int q)
{
	quadrant = q;
}

int SimulationCore::getDimensions()
{

	return dimensions;
}

int SimulationCore::displacementDimensions()
{
	int displacements = 0;
	
	if(x_index >= 0)
		++displacements;
	if(y_index >= 0)
		++displacements;
	if(z_index >= 0)
		++displacements;

	return displacements;
}

PluginManager* SimulationCore::pluginManager()
{
	return plugin_man;
}


void SimulationCore::runExperimentManager()
{
	exp_man->init();
	exp_man->run();
	delete pSimulationCore;
	exit(0);
}

void SimulationCore::runPluginManager()
{
	plugin_man->init();
	plugin_man->run();
	delete pSimulationCore;
	exit(0);
}

void SimulationCore::abort(string msg)
{
	cerr<<msg<<"\nAborting..."<<endl;
	delete this;
	exit(2);
}
/*------------------------------------------------------------------------------*/
/* DEBUG- OR USERINFO								*/
/*------------------------------------------------------------------------------*/
extern void Debug( char* format, ... )
{
#ifndef NDEBUG
	va_list args;
	va_start( args, format );
	fprintf( stderr, "[debug] " );
	vfprintf( stderr, format, args );
	fprintf( stderr, "\n" );
	va_end( args );
	fflush( stderr );
#endif /* not defined(NDEBUG) */
}
