/***************************************************************************
 * File:        ./crusde_api.c
 * Author:      Ronni Grapenthin, NORVULK & HU-BERLIN
 * Created:     20.02.2007
 * Licence:     GPLv2
 *
 * #########################################################################
 *
 * CrusDe, simulation framework for crustal deformation studies
 * Copyright (C) 2007 Ronni Grapenthin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * -------------------------------------------------------------------------
 * Description: Implements the Application Programming Interface for
 *		the green's functions modeling tool.
 ****************************************************************************/

#include <iostream>
#include <map>
#include <string>
#include "crusde_api.h"
#include "SimulationCore.h"
#include "DataOutPlugin.h"
#include "ConvFactorPlugin.h"
#include "LoadHistoryPlugin.h"

using namespace std;

/* forward declaration of private functions */
void print_params(multimap<string, double*> m);


/* PARAMETER REGISTRATION */

extern "C" void crusde_register_param(double* param, const char* name, PluginCategory category)
{
printf("crusde_register_param called for: %s\n", name);
	if(category == LOAD_PLUGIN){
		SimulationCore::s_params.load.insert( pair< string, double*>(string(name), param) );
	} 
	else if(category == KERNEL_PLUGIN){
		SimulationCore::s_params.kernel.insert( pair< string, double*>(string(name), param) );
	} 
	else if(category == GREEN_PLUGIN){
		SimulationCore::s_params.green.insert( pair< string, double*>(string(name), param) );
	} 
	else if(category == DATAOUT_PLUGIN){
		SimulationCore::s_params.datahandler.insert( pair< string, double*>(string(name), param) );
	} 
	else if(category == POSTPROCESS_PLUGIN){
		SimulationCore::s_params.postprocessor.insert( pair< string, double*>(string(name), param) );
	}
	else if(category == LOADHISTORY_PLUGIN){
		SimulationCore::s_params.loadhistory.insert( pair< string, double*>(string(name), param) );
	} else {
		cerr<<"in crusde_register_param: unknown category: "<<category<<endl;
	}
}

/* OUTPUT FIELD REGISTRATION */

extern "C" void crusde_register_output_field(int* position, FieldName field)
{
    return SimulationCore::instance()->registerOutputField(position, field);
}

/* PARAMETER PRINTING */

extern "C" void crusde_print_green_params()
{
	print_params(SimulationCore::s_params.green);
}

extern "C" void crusde_print_load_params()
{
	print_params(SimulationCore::s_params.load);
}

extern "C" void crusde_print_kernel_params()
{
	print_params(SimulationCore::s_params.kernel);
}

void print_params(multimap<string , double*> m)
{
  	multimap<string , double*>::iterator iter = m.begin();
	
	while( iter != m.end()) 
	{
		printf("%s = %f\n", iter->first.c_str(), *(iter->second));
		++iter;
 	}
}

int crusde_get_green_at(double **result, int x, int y)
{
    return SimulationCore::instance()->greenPlugin()->getValueArrayAt(result,x,y);
}

double crusde_get_load_at(int x, int y, int t)
{
    return SimulationCore::instance()->loadPlugin()->getValueAt(x,y,t);
}


int crusde_get_size_x()
{
    return SimulationCore::instance()->sizeX();
}

int crusde_get_size_y()
{
    return SimulationCore::instance()->sizeY();
}

int crusde_get_size_t()
{
    return SimulationCore::instance()->sizeT();
}

int crusde_get_gridsize()
{
    return SimulationCore::instance()->gridSize();
}

int crusde_get_dimensions()
{
    return SimulationCore::instance()->getDimensions();
}

int crusde_get_displacement_dimensions()
{
    return SimulationCore::instance()->displacementDimensions();
}

int crusde_get_min_x()
{
    return SimulationCore::instance()->minX();
}

int crusde_get_min_y()
{
    return SimulationCore::instance()->minY();
}

extern "C" const char* crusde_get_load_file()
{
	return SimulationCore::instance()->loadFile();
}

extern "C" void crusde_set_result(double **result)
{
    SimulationCore::instance()->setModelData(result);
}

extern "C" double** crusde_get_result()
{
    return SimulationCore::instance()->getModelData();
}

extern "C" void crusde_set_quadrant(int q)
{
    SimulationCore::instance()->setQuadrant(q);
}

extern "C" int crusde_get_quadrant()
{
    return SimulationCore::instance()->getQuadrant();
}

extern "C" const char* crusde_get_out_file()
{
    return SimulationCore::instance()->outFile();
}

extern "C" void crusde_exit(int exitcode)
{
    SimulationCore::instance()->terminate();
    exit(exitcode);
}

extern "C" int crusde_load_history_exists()
{
    return SimulationCore::instance()->loadHistoryPlugin()->isLoaded() ? 1 : 0;
}

extern "C" int crusde_model_time()
{
    return SimulationCore::instance()->modelTime();
}

extern "C" int crusde_stepsize()
{
    return SimulationCore::instance()->stepSize();
}

extern "C" int crusde_get_x_index()
{
    return SimulationCore::instance()->xIndex();
}

extern "C" int crusde_get_y_index()
{
    return SimulationCore::instance()->yIndex();
}

extern "C" int crusde_get_z_index()
{
    return SimulationCore::instance()->zIndex();
}

extern "C" double crusde_constrain_load_height(double load, int x, int y, int t)
{
    if(t!=NO_TIME && SimulationCore::instance()->loadHistoryPlugin()->isLoaded() )
    {
    	return SimulationCore::instance()->loadHistoryPlugin()->constrainLoadHeight(load, x, y, t);
    }
    else
    {
    	return load;
    }
}

extern "C" green_exec_function crusde_request_green_plugin(char* plugin) 
{
	try
	{
		return SimulationCore::instance()->addGreenPlugin( string(plugin) );
	}
	catch( runtime_error e)
	{
		SimulationCore::instance()->abort(e.what());
	}
	
	return NULL;
}

extern "C" load_exec_function crusde_request_load_plugin(char* plugin) 
{
	try
	{
		return SimulationCore::instance()->addLoadPlugin( string(plugin) );
	}
	catch( runtime_error e)
	{
		SimulationCore::instance()->abort(e.what());
	}
	
	return NULL;

}

extern "C" run_function crusde_request_kernel_plugin(char* plugin)
{
	try
	{
		return SimulationCore::instance()->addKernelPlugin( string(plugin) );
	}
	catch( runtime_error e)
	{
		SimulationCore::instance()->abort(e.what());
	}
	
	return NULL;
	
}

extern "C" run_function crusde_request_postprocessor_plugin(char* plugin)
{
	try
	{
		return SimulationCore::instance()->addPostprocessorPlugin( string(plugin) );
	}
	catch( runtime_error e)
	{
		SimulationCore::instance()->abort(e.what());
	}
	
	return NULL;
	
}

extern "C" loadhistory_exec_function crusde_request_loadhistory_plugin(char* plugin) 
{
	try
	{
		return SimulationCore::instance()->addLoadHistoryPlugin( string(plugin) );
	}
	catch( runtime_error e)
	{
		SimulationCore::instance()->abort(e.what());
	}
	
	return NULL;

}
