/***************************************************************************
 * File:        ./plugins/green/disk_load.c
 * Author:      Ronni Grapenthin, NORVULK & HU-BERLIN
 * Created:     21.02.2007
 * Licence:     GPL
 ****************************************************************************/

/** 
 *  @ingroup LoadFunction
 *  @ingroup Plugin 
 **/

/*@{*/
/** \file irregular_load.c
 * 
 * Read load from a file that must be in the following format:
 *
 *		Lon Lat Height
 *
 * with Lon and Lat being integers in Lambert coordinates and Height determining the load height 
 * at the point (Lon,Lat).\
 *
 */
/*@}*/

#include <stdio.h>
#include <math.h>

#include "crusde_api.h"

/*load command line parameters*/
double rho;		/*!< Density of the load	[kg/m^3]	*/
const char* load_file;	/*!< Name of the file that contains irregular load*/

/* internals */
int dS;			/*!< Area around point P(x,y)	[m^2]		*/
double rho_dS_const;	/*!< rho_dS_const = rho * dS	[kg/m] 		 */
double *load_array = NULL;
int nx=-1, ny=-1;

extern const char* get_name() 	     { return "irregular load"; }
extern const char* get_version()     { return "0.1"; }
extern const char* get_authors()     { return "ronni grapenthin"; }
extern const char* get_description() { 
	return "Read load from a file that must be in the following format:<br/><br/> \
	<center><code>Lon Lat Height</code></center><br/><br/> \
	with <code>Lon</code> and <code>Lat</code> being integers in Lambert coordinates \
	and <code>Height</code> determining the load height at point (lon,lat).\
	In case a load history is defined for a simulation it is used to constrain the height for\
	the current time step.\
	"; }
extern const PluginCategory get_category() { return LOAD_PLUGIN; }

/*!empty*/
extern void request_plugins()
{
//	disk_load = crusde_request_load_plugin("disk load");
//	disk_load2 = crusde_request_load_plugin("disk load");
}

/*!empty*/
extern int register_output_fields(){};

/*!empty*/
extern void run(){}

/*!frees load array that was read from file*/
extern void clear()
{
	free(load_array);
}

void gmt_print_double_array(int m, int n, double* M){
   int k=-1, l=-1;
   while(++k<n){
	l = -1;
   	while(++l<m){
	   	printf("%d %d %f\n", k+1, l+1, M[k*n+l]);
	}
   }

}

void print_double_array(int size_x, int size_y, double* M){
   int y=-1, x=-1;

//   printf("----------------------------------------------------\n");
   while(++y < size_y){
	x = -1;
  	while(++x < size_x){	
		if(y+1==size_y)	
	   		printf("%.1f ", M[y+size_y*x]);
		else
			printf("%.1f, ", M[y+size_y*x]);
	}
   	printf("\n");
   }
//   printf("----------------------------------------------------\n");

}


//! Initialize members that depend on registered values. 
/*! This function <b>must not</b> be called before register_parameter() unless none of the necessary values depends on 
 *  parameters provided by the user, which are only set after they have been registered. This function is called
 *  some time <b>after</b> register_parameter().
 */
extern void init()
{
	FILE *fi;
	int first_x=-1;
	int prev_x=-1, prev_y=-1;
	int dimX=0, dimY=0;
	int inc_x, inc_y;
	boolean x_set=false, y_set=false;
	float xg,yg,zg;       		/* buffers to store values gotten from 
				   the input file.			*/

	int x=0,y=0;
	int x_min = crusde_get_min_x();
	int y_min = crusde_get_min_y();
	int gridsize = crusde_get_gridsize();
	
	/* get name of the file holding the irregular load*/
	load_file=crusde_get_load_file();
	
	/*open load file*/
	fi = fopen( load_file, "r" );

	/*file exists?*/	
	if(fi == NULL)
	{
		perror(load_file);
		exit(1);
	}

	/*other problems with file?*/
	if(ferror(fi) != 0)
	{
		perror(load_file);
		exit(1);
	}

	/*determine array dimensions and increment from load file*/
//	while(fscanf(fi, "%f%f%f", &yg,&xg,&zg) != EOF)
//	{	
//		/*increment data counter*/
//		if(!x_set){
//		
//			/* if x column repeats for the first time set all possible values*/
//			if(first_x == (int) xg && nx == -1)
//			{
//				nx    = dimX;
//				inc_y = (int) yg - prev_y;
//				x_set = true;
//			}
//			else
//			{	/*inc_x can be calculated at any time but not when first_x==xg --> values are not neighbors*/
//				inc_x = (int) xg - prev_x;
//			}
//
//			/* save first coordinate for later comparisons*/
//			if(first_x == -1)
//			{
//				first_x = (int) xg;
//			}
//			/* buffer previous values for increment determination */
//			prev_x = (int) xg;
//			prev_y = (int) yg;
//			++dimX;
//	
//		}
//
//		++dimY;
//		
//	}/*while ! EOF*/
//
//	/*y dimension == number of lines with %f%f%f format over x dimension*/
//	ny    = (int) dimY/nx;
//	dS = ;

	nx = crusde_get_size_x();
	ny = crusde_get_size_y();
	dS = gridsize*gridsize;

//printf("nx = %d, ny =%d, dS =%d\n", nx, ny, dS);

	load_array = (double*) malloc (sizeof(double) * nx * ny);
			
	/*reset filepointer*/
	rewind(fi);
	
	inc_x = -1;
	
	while(++inc_x < nx*ny)
	{
		load_array[inc_x] = 0.0;
	}
	/*read values, copy to local array*/
	while(fscanf(fi, "%f%f%f", &xg,&yg,&zg) != EOF){	
		x = (int) ( ( xg - x_min) / gridsize );
		y = (int) ( ( yg - y_min) / gridsize );
		load_array[y+ny*x] = zg;
	}/*while ! EOF*/

	rho_dS_const = rho * (double) dS;
//printf("rho = %f, rho_dS_const  = %f \n", rho, rho_dS_const );
	
//	gmt_print_double_array(nx, ny, load_array);
//	print_double_array(nx, ny, load_array);

}



//! Register parameters this load function claims from the input.
/*! This function calls register_load_param() defined in crusde_api.h to register 
 *  references to parameters this Green's function will need to operate properly.
 *  In case an XML is used to configure the experiment, the reference to this parameter will be identified by 
 *  the string passed as second argument to register_green_param().
 *
 *  This function registers the references in the following order:
 * 
 *  -# rho, XML config identifier is "rho", SI-unit [kg/m^3]
 *
 *  @see register_load_param()
 *  @see register_parameter() (temp_plugin.c.tmp)
 */
extern void register_parameter()
{
   /* tell main program about parameters we claim from input */
   crusde_register_param(&rho, "rho", get_category());
}

//! Returns the Load of a disk at Point (x,y) at time t.
/*! Computes the euclidean distance of Point (x,y) to (center_x, center_y). 
 *  If this distance is less or equal to disk_radius the load will be returned, zero otherwise.
 *  
 * @param x The x-Coordinate of the wanted value.
 * @param y The y-Coordinate of the wanted value.
 * @param t The time-Coordinate of the wanted value (set to NO_TIME in non-temporal modelling environment).
 * 
 * @return The load at Point (x,y,t).
 */
extern double get_value_at(int x, int y, int t)
{
	if(crusde_load_history_exists() == true)
	{
		/*do time dependent load modeling if requested*/
		return ( crusde_constrain_load_height(load_array[y+ny*x], x, y, t) * rho_dS_const);
	}
	else
	{
		return (load_array[y+ny*x] * rho_dS_const);
	}
}
