/***************************************************************************
 * File:        ./plugins/load_history/sinusoidal.c
 * Author:      Ronni Grapenthin, NORVULK & HU-BERLIN
 * Created:     29.04.2007
 * Licence:     GPL
 ****************************************************************************/

/** 
 *  @defgroup LoadHistory Load History Functions
 *  @ingroup Plugin 
 **/

/*@{*/
/** \file sinusoidal.c
 * Sinusoidal load history that calculates a load height depending on the
 * current time step (t), a period length (p) and the timestep when the load is supposed to
 * be at maximum (<i>d<sub>m</sub></i>):
 *
 * \f[
 *	h(t) = \frac{h_m}{2} \left[ 1 + cos( \frac{2\,\pi}{p} (t - d_m) ) \right]
 * \f]
 * with <i>h<sub>m</sub></i> being the maximum load height.
 */
/*@}*/

#include <stdio.h>
#include <math.h>

#include "crusde_api.h"

/*load command line parameters*/
double d_max;		/*!< day of maximum load	[d]		*/
double period_length;	/*!< number of days in a cycle	[d]		*/

extern const char* get_name() 	     		{ return "sinusoidal"; }
extern const char* get_version()     		{ return "0.1"; }
extern const char* get_authors()     		{ return "ronni grapenthin"; }
extern const PluginCategory get_category() 	{ return LOADHISTORY_PLUGIN; }
extern const char* get_description() 		{ 
	return "Sinusoidal load history that calculates a load height depending on the\
	current time step (<i>t</i>), a period length (<i>p</i>) and the timestep when the load is supposed to \
	be at maximum (<i>d<sub>m</sub></i>):<br/><br/>\
	h(t) =  h<sub>m</sub> / 2 [ 1 + cos( 2*&pi; / p * (t-d<sub>m</sub>) ) ]\
	<br/><br/>\
	with <i>h<sub>m</sub></i> being the maximum load height.\
	"; }

/*! empty*/
extern void request_plugins(){}
/*! empty*/
extern int register_output_fields(){};
/*! empty*/
extern void run(){}
/*! empty*/
extern void clear(){}
/*! empty*/
extern void init(){}



/*! Register parameters this load function claims from the input.*/
extern void register_parameter()
{
   /* tell main program about parameters we claim from input */
   crusde_register_param(&d_max, "peak", get_category());
   crusde_register_param(&period_length, "period_length", get_category());
}

/*! Returns the constrained load depending on the current timestep at point (x,y)*/
extern double constrain_load_height(double h_max, int x, int y, int t)
{
	return (( h_max + h_max * cos( PI*2 / period_length * (t-d_max) ) ) / 2 )  ;
}
